from kestrel.datasource.interface import MODULE_PREFIX, AbstractDataSourceInterface
from kestrel.datasource.retstruct import ReturnFromFile, ReturnFromStore
from kestrel.datasource.manager import DataSourceManager
